from typing import List, Optional

from . import BaseSchema


class UrlView(BaseSchema):
    url: str


class ManifestType(BaseSchema):
    id: str
    name: str


class PreviewManifest(BaseSchema):
    url: str
    width: int
    height: int


class SuggestServiceManifestEntry(BaseSchema):
    service_url: Optional[str] = None
    service_path: Optional[str] = None
    flyout_service_url: Optional[str] = None
    flyout_service_path: Optional[str] = None


class SuggestServiceManifest(BaseSchema):
    entity: Optional[SuggestServiceManifestEntry] = None
    property: Optional[SuggestServiceManifestEntry] = None
    type: Optional[SuggestServiceManifestEntry] = None


class Manifest(BaseSchema):
    versions: List[str]
    name: str
    identifier_space: str
    schema_space: str
    documentation: Optional[str] = None
    service_version: Optional[str] = None
    logo: Optional[str] = None
    view: Optional[UrlView] = None
    feature_view: Optional[UrlView] = None
    default_types: Optional[List[ManifestType]] = None
    preview: Optional[PreviewManifest] = None
    suggest: Optional[SuggestServiceManifest] = None
    batch_size: Optional[int] = None
    # TODO:
    #  - extend is not yet supported
    #  - authentication is not yet supported



from typing import Any, Dict, List, Optional

import httpx

from .models.manifest import Manifest
from .models.reconciliation import (
    BatchReconciliationQuery,
    BatchReconciliationResult,
    ReconciliationQuery,
    ReconciliationResult,
)


class ReconciliationService:
    _base_url: str
    _manifest: Manifest
    _async_client: Optional[httpx.AsyncClient] = None
    _sync_client: Optional[httpx.Client] = None
    _httpx_args: Dict[str, Any]

    def __init__(self, base_url: str, manifest: Manifest, httpx_args: Dict[str, Any] = {}):
        self._base_url = base_url
        self._manifest = manifest
        self._httpx_args = httpx_args

    def _get_sync_client(self) -> httpx.Client:
        if self._sync_client is None:
            self._sync_client = httpx.Client(**self._httpx_args)
        return self._sync_client

    async def _get_async_client(self) -> httpx.AsyncClient:
        if self._async_client is None:
            self._async_client = httpx.AsyncClient(**self._httpx_args)
        return self._async_client

    def reconcile(self, query: ReconciliationQuery) -> ReconciliationResult:
        return self.reconcile_batch([query])[0]

    def reconcile_batch(self, query: List[ReconciliationQuery]) -> List[ReconciliationResult]:
        queries = {f"q{i}": q for i, q in enumerate(query)}
        client = self._get_sync_client()
        batch_query = BatchReconciliationQuery.model_validate(queries)
        batch_query_json = batch_query.model_dump_json(exclude_none=True)
        resp = client.post(self._base_url, data={"queries": batch_query_json})
        if resp.status_code != 200:
            raise Exception(f"Could not reconcile batch: {resp.status_code} {resp.text}")
        batch_result = BatchReconciliationResult.model_validate(resp.json())
        out = []
        for i in range(len(query)):
            if f"q{i}" not in batch_result.root:
                raise Exception(f"Missing result for query q{i}")
            out.append(batch_result.root[f"q{i}"])
        return out


def build_reconciliation_service(base_url: str, *, httpx_args: Dict[str, Any] = {}) -> "ReconciliationService":
    client = httpx.Client(**httpx_args)

    manifest_resp = client.get(base_url)
    if manifest_resp.status_code != 200:
        raise Exception(f"Could not fetch manifest from {base_url}: {manifest_resp.status_code} {manifest_resp.text}")

    if manifest_resp.headers.get("content-type") != "application/json":
        raise Exception(
            f"Manifest from {base_url} has unexpected content type: {manifest_resp.headers.get('content-type')}"
        )

    manifest = Manifest.model_validate(manifest_resp.json())

    if "0.2" not in manifest.versions:
        raise Exception(f"Manifest from {base_url} has no version 0.2. Only supports versions {manifest.versions}")

    return ReconciliationService(base_url, manifest)



from typing import Dict, List, Literal, Optional, Union

from pydantic import RootModel
from pydantic.types import PositiveInt

from . import BaseSchema


class CompositePropertyValue(BaseSchema):
    id: str
    name: Optional[str] = None


PropertyValue = Union[str, float, int, bool, CompositePropertyValue]


class ReconciliationQueryProperty(BaseSchema):
    pid: str
    v: Union[List[PropertyValue], PropertyValue]


class ReconciliationQuery(BaseSchema):
    query: str
    type: Optional[Union[str, List[str]]] = None
    limit: Optional[PositiveInt] = None
    properties: Optional[List[ReconciliationQueryProperty]] = None
    type_strict: Optional[Literal["should", "all", "any"]] = None


class BatchReconciliationQuery(RootModel):
    root: Dict[str, ReconciliationQuery]


class ReconciliationFeature(BaseSchema):
    id: str
    value: Union[bool, float, int]


class ReconciliationType(BaseSchema):
    id: str
    name: Optional[str] = None


class ReconciliationCandidate(BaseSchema):
    id: str
    name: str
    score: float
    features: Optional[List[ReconciliationFeature]] = None
    match: Optional[bool] = None
    type: Optional[List[Union[str, ReconciliationType]]] = None


class ReconciliationResult(BaseSchema):
    result: List[ReconciliationCandidate]


class BatchReconciliationResult(RootModel):
    root: Dict[str, ReconciliationResult]



from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class BaseSchema(BaseModel):
    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True)
